package djmil.cordacheckers.gameboard;

import java.time.Duration;
import java.time.Instant;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import djmil.cordacheckers.checkers.Move;
import djmil.cordacheckers.contracts.GameCommand;
import djmil.cordacheckers.gameresult.GameResultCommiter;
import djmil.cordacheckers.gamestate.CommitTrx;
import djmil.cordacheckers.gamestate.FlowResponce;
import djmil.cordacheckers.gamestate.GetFlow;
import djmil.cordacheckers.gamestate.View;
import djmil.cordacheckers.gamestate.ViewBuilder;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameState;
import net.corda.v5.application.flows.ClientRequestBody;
import net.corda.v5.application.flows.ClientStartableFlow;
import net.corda.v5.application.flows.CordaInject;
import net.corda.v5.application.flows.FlowEngine;
import net.corda.v5.application.marshalling.JsonMarshallingService;
import net.corda.v5.application.membership.MemberLookup;
import net.corda.v5.base.annotations.Suspendable;
import net.corda.v5.base.types.MemberX500Name;
import net.corda.v5.crypto.SecureHash;
import net.corda.v5.ledger.utxo.StateAndRef;
import net.corda.v5.ledger.utxo.UtxoLedgerService;
import net.corda.v5.ledger.utxo.transaction.UtxoSignedTransaction;

public class MoveFlow implements ClientStartableFlow{

     private final static Logger log = LoggerFactory.getLogger(MoveFlow.class);

     @CordaInject
     public JsonMarshallingService jsonMarshallingService;

     @CordaInject
     public UtxoLedgerService utxoLedgerService;

     @CordaInject
     public FlowEngine flowEngine;

     @CordaInject
     public MemberLookup memberLookup;

     @Suspendable
     @Override
     public String call(ClientRequestBody requestBody) {
          SecureHash utxoTrxId = null;

          try {
               final MoveFlowArgs args = requestBody.getRequestBodyAs(jsonMarshallingService, MoveFlowArgs.class);
               
               final GameCommand command = new GameCommand(args.move);

               final StateAndRef<GameState> currentGameSar = this.flowEngine
                    .subFlow(new GetFlow(args.gameUuid));
               final GameBoardState currenGameBoardState = (GameBoardState)currentGameSar.getState().getContractState();

               final GameBoardState newGameBoard = new GameBoardState(
                    currenGameBoardState, 
                    args.move,
                    args.message);

               final UtxoSignedTransaction moveTrx = utxoLedgerService.createTransactionBuilder()
                    .addCommand(command)
                    .addInputState(currentGameSar.getRef())
                    .addOutputState(newGameBoard)
                    .addSignatories(newGameBoard.getParticipants())
                    .setNotary(currentGameSar.getState().getNotaryName())
                    .setTimeWindowUntil(Instant.now().plusMillis(Duration.ofDays(1).toMillis()))
                    .toSignedTransaction();

               utxoTrxId = this.flowEngine
                    .subFlow(new CommitTrx(moveTrx, currenGameBoardState.getIdelPlayerName()));

               if (amIwon(newGameBoard)) {
                    log.info("Opponent has no possible moves. Claim victory!");
                    utxoTrxId = this.flowEngine
                         .subFlow(new GameResultCommiter(
                                   newGameBoard.getGameUuid(), 
                                   new GameCommand(GameCommand.Action.CLAIM_VICTORY)));
               }

               final View gameStateView = this.flowEngine
                    .subFlow(new ViewBuilder(utxoTrxId));

               return new FlowResponce(gameStateView, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          }
          catch (Move.Exception e) {
               return new FlowResponce(e.getMessage(), utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          }
          catch (Exception e) {
               log.warn(requestBody + " [ERROR] " +e.toString());
               return new FlowResponce(e, utxoTrxId)
                    .toJsonEncodedString(jsonMarshallingService);
          }
     }

     @Suspendable
     boolean amIwon(GameBoardState gameBoard) {
          final MemberX500Name myName = memberLookup.myInfo().getName();
          
          if (gameBoard.getActivePlayerName().compareTo(myName) == 0 || 
              gameBoard.getIdelPlayerName().compareTo(myName) != 0)
               return false; // a bit of overkill check, but you have to finish your move first

          final var possibleOponnentMoves = Move.getPossibleMoves(gameBoard.getBoard(), gameBoard.getActiveColor());

          return possibleOponnentMoves.isEmpty(); // true, if my opponent has no possible moves
     }

}
